#ifndef SATGEO_H
#define SATGEO_H
#include <iostream>
#include "orbitgeo.h"


/**
La classe SatGeo est une classe qui permet de definir les position, attitude 
d'un satellite et les vitesses associees a tout instant.
<P>
Dans la designation des composantes, les indices _n, _c, _g et _l sont relatifs aux
reperes suivants :
<UL>
<LI> Rn : repere lie a l'orbite avec x selon ligne des Noeuds et z selon H cinetique
<LI> Rc : repere equatorial Celeste (z vers le pole Nord et x vers Gamma J2000)
<LI> Rg : repere geographique (xOz contient Greenwich)
<LI> Rl : MON repere orbital Local (x~//Vit, y//Hcin, z//Terre->Sat )
<LI> Rs : Le repere lie au satellite
</UL>
<P>
Les unites pour toutes les variables sont en S.I. L'origine des temps (en secondes) est
le 1er janvier 2000 a midi a Greenwich (J2000.0).
@author Michel Llibre - ONERA/DCSD.
 */

typedef	struct {
    double a;
    double ex;
    double ey;
    double i0;
    double gdOm0;
    double pso;
    double t0;
    char nom[80];
} CnesParameters;

class TwoLE;

class SatGeo : public OrbitGeo
{
private :	 
    double n0dots2; 	// 1/2 derivee premiere
    double n0ddots6;   // et 1/6 derivee seconde du mouv. moyen
    double v3_hcin[3];
    double v3_Zloc[3];
    double v3_Vs_n[3];
    double m3_CH[9];
    double m3_NL[9];
    void calFlagND();

public :
    char nom[80];
    double t0;          // Date des elements orbitaux (s/J2000)
    double M0;          // Anomalie moyenne a l'instant t0
    int flag_ND;        // Flag vrai si noeud descendant cote Soleil
    double tcou;        // Instant courant des parametres de position
    double dist;        // Distance satellite - centre Terre en metres
    double grand_Omega; // Ascension droite du noeud ascendant en radian
    double petit_omega; // Argument perigee en radian
    double anoMoy_M;    // Anomalie moyenne
    double anoVraie_v;  // Anomalie vrai
    double pso;         // Argument de latitude du satellite (arg perigee+ano_vraie) en rd
    double ascen_droite;// Ascension droite du satellite dans Rc en radian
    double declinaison; // Declinaison du satellite dans Rc en radian
    double longitude;   // Longitude du satellite dans Rg en radian
    double latitude;    // Latitude du satellite dans Rg en radian
    double v3_Vs_c[3]; // Vitesse translation du satellite dans Rc
    double m3_CL[9];   // Matrice attitude du repere orbital local Rl / Rc
    double m3_GL[9];   // Matrice attitude du repere orbital local Rl / Rg
    double v3_Oml_c[3];// Vitesse rotation du repere orbital local / Rc
    double m3_CN[9];   // Attitude repere plan orbital Rn / Rc
    double v3_TS_c[3]; // Terre-SatGeo dans Rc

/** Constructeur par défaut */
SatGeo(){;}

/** Construction d'un satellite a partir de ses elements orbitaux
 *  en unites S.I.
  @param nom	nom du satellite,
  @param n0	mouvement moyen(radians/s)
  @param e	excentricite
  @param i0	inclinaison de l'orbite (radians)
  @param om0	argument du perigee a la date t0 (radians)
  @param Om0	ascension droite du noeud ascendant a la date t0 (radians)
  @param M0	anomalie moyenne a la date t0 (radians)
  @param t0	date de mesure des elements orbitaux (secondes relatives J2000)
 */
    SatGeo(char *name, double n0, double e, double i0, double om0, double Om0, double M0, double t0);
    void set(char *name, double n0, double e, double i0, double om0, double Om0, double M0, double t0);

/** Construction d'un satellite a partir de ses elements orbitaux modifiés en S.I., de type CNES
    Dans la liste des arguements, le nom est placé en dernier pour avoir une signature différente
    du contruteucr de base.
  @param a   	demi grand axe (m)
  @param ex	excentricite*cos(omega)
  @param ey   	excentricite*sin(omega)
  @param incl  	inclinaison (radians)
  @param Om0  	ascension droite du noeud ascendant a la date t0 (radians)
  @param pso	argument du satellite a t0 (radians)
  @param t0  	date de mesure des elements orbitaux (secondes a partir de J2000)
  @param nom	nom du satellite,
 */
    SatGeo(CnesParameters *p);
    SatGeo(double a, double ex, double ey, double i0, double Om0, double pso, double t0, char * name);
    void set(double a, double ex, double ey, double i0, double Om0, double pso, double t0, char *name);

/** Creation d'un satellite a partir de ses elements orbitaux
 * lus dans un fichier. Les parametres lus sont dans l'ordre :
 * - le nom du satellite,
 * - la date t0 de mesure des elements orbitaux (lu en jours a partir de J2000)
 * - l'inclinaison i0 de l'orbite (lu en degres),
 * - l'ascension droite du noeud ascendant Om0 a la date t0 (lu en degres),
 * - l'excentricite,
 * - l'argument du perigee om0 a la date t0 (lu en degres),
 * - l'anomalie moyenne M0 a la date t0 (lue en degres),
 * - le mouvement moyen n0 (lu en tours/jour).
 * @param filename  Nom du fichier contenant les elements.
 */
    SatGeo(char *filename);
    int set(char *filename);

/**
 * Construction satellite a partir de ses TLE
 * @param tle
 */
    SatGeo(TwoLE *tle);
    void set(TwoLE *tle);

/** Creation d'un satellite a partir de sa position-vitesse :
  @param nom	nom du satellite,
  @param t0	date de mesure des elements (s a partir de J2000)
  @param v3d_TS_c	Vect3 Position satellite dans Rc
  @param v3d_Vs_c	Vect3 Vitesse satellite dans Rc
  @return new SatGeo cree.
 */
    SatGeo(char * name, double t0, double* v3d_TS_c, double* v3d_Vs_c);
    int set(char * name, double t0, double *v3d_TS_c, double *v3d_Vs_c);

/** Creation d'une orbite heliosynchrone circulaire :
  @param nom   Nom du satellite
  @param Q     Nombre de jours de la periode de repetitivite
  @param K     Nombre de revolutions dans la periode de repetitivite
  @param SNA_h Ecart horaire decimal Soleil - noeud ascendant
  @param jNA   date de passage au noeud ascendant(jour decimal/J2000)
 */
    SatGeo(char * name, int Q, int K, double SNA_h, double jNA);
    void set(char * name, int Q, int K, double SNA_h, double jNA);

/** Mise en coherence des elements orbitaux d'un satellite a partir
 * d'un lot de ses parametres en unites S.I.
  @param n0		mouvement moyen(radians/s)
  @param e			excentricite
  @param i0		inclinaison de l'orbite (radians)
  @param om0		argument du perigee a la date t0 (radians)
  @param Om0		ascension droite du noeud ascendant a la date t0 (radians)
  @param M0		anomalie moyenne a la date t0 (radians)
  @param t0		date de mesure des elements orbitaux (secondes relatives J2000)
  @param tcou		date de calcul de l'etat (secondes relatives J2000)
 */
    void update(double n0, double e, double i0,
                     double om0, double Om0, double M0_, double t0_, double tcou);


/** Calcul de l'etat satellite (cad de Ccn, TS_c, Vs_c, Ccl, Oml_c)
 @param t    date de calcul de l'etat en s a partir de J2000.0
*/
    void etat(double t, double precision);

/** Calcul de l'etat satellite (cad de Ccn, TS_c, Vs_c, Ccl, Oml_c)
 * Methode pour effectuer des test. Elle n'est pas validee.
 * Elle ne met pas à jour tcou de maniere a ne pas empecher le calcul
 * ulterieur par la methode etat(t) standard.
  @param t    date de calcul de l'etat en s a partir de J2000.0
 */
    void etatSGP(double t, double precision);

/**
 * Impression paramètres orbitaux
 */
    void print();
/**
 * Impression de l'etat du satellite
 */
    void printEtat();

    static void test();
};
#endif
