
#include "univglwidget.h"

#include <QtGui>
#include <QAction>
#include <QFileDialog>
#include "../cinespace.h"
#include <math.h>
#include "terre.h"
#include "reperes.h"
#include "lune.h"
#include "../grafutils/qtglutils.h"
#include "../mathutil/rotation.h"
#include "../mathutil/sp3.h"
#include "planets.h"

static double gOmega[8] = {0.843, 1.3387, 0., 0.8657, 1.7523, 1.984, 1.2915,   2.3};
static double inclin[8] = {0.12217, 0.05934, 0., 0.03229, 0.02278, 0.04335, 0.01344, 0.03089};
static double pomega[8] = {0.50824, 0.95731, 1.79594, 5.00037, 4.78220, 5.91318, 1.71042, 4.81711};
static double demgaxe[8] = {0.387, 0.723, 1., 1.524, 5.202, 9.555, 19.2, 30.1};
static double excen[8] = {0.206, 0.0068, 0.0167, 0.093, 0.048, 0.055, 0.046, 0.0086};
static GLdouble colOrb[8][3] = {{0.2, 0.2f, 0.2f},
                                {1.0, 1.0, 0.0},
                                {0.47, 0.71, 1.0},
                                {1.0, 0.0, 0.0},
                                {7.0, 7.0, 0.2},
                                {1.0, 1.0, 0.0},
                                {0.0, 1.0, 0.0},
                                {0.0, 0.0, 1.0}};
								
/*
 *  Fabrication d'une DisplayListe openGL d'affichage d'une orbite
 *  Entrée :
 *     a : demi-grand axe de l'orbite
 *     e : excentricité de l'orbite
 *     I : inclinaison de l'orbite
 *     O : longitude du noeud ascendant de l'orbite
 *     w : argument du périgée
 *  Retour : numéro de la DisplayList générée.
 *
*/
static GLuint glNewOrbite(double a, double e, double I, double O, double w, int nbpt)
{
    int i;
    double anoE, cE, sE, cw = cos(w), sw = sin(w);
    double cI = cos(I), sI = sin(I), cO = cos(O), sO = sin(O), sOcI = sO*cI, cOcI = cO*cI;
    double c = a*e, b = a*sqrt(1.-e*e);
    double x1, y1, x, y, X, Y, Z;
    GLuint dl_orbite;
    dl_orbite = glGenLists(1);                      // Nom de l'objet que l'on fabrique
    glNewList(dl_orbite, GL_COMPILE);
    glBegin(GL_LINE_LOOP);
    for(i = 0; i < nbpt; i++)
    {
        anoE = i*2.*M_PI/nbpt; cE = cos(anoE), sE = sin(anoE);
        x1 = a*(cE - e); y1 = b*sE;
        x = x1*cw - y1*sw; y = x1*sw + y1*cw;
        X = x*cO - y*sOcI;
        Y = x*sO + y*cOcI;
        Z = y*sI;
        glVertex3d(X,Y,Z);
    }
    glEnd();
    glEndList();                                     // Fin de définition de l'objet
    return dl_orbite;
}
								


// Quelques intensités pour les éclairages openGL
static GLfloat eteinte[] = {0.0f, 0.0f, 0.0f, 1.0f};
static GLfloat tenue[] = {0.1f, 0.1f, 0.1f, 1.0f};
static GLfloat faible[] = {0.2f, 0.2f, 0.2f, 1.0f};
static GLfloat moyenne[] = {0.5f, 0.5f, 0.5f, 1.0f};
static GLfloat forte[] = {0.8f, 0.8f, 0.8f, 1.0f};
static GLfloat puissante[] = {1.0f, 1.0f, 1.0f, 1.0f};


// La classe d'affichage du Widget openGL de la scene spatiale
// -----------------------------------------------------------
UnivGLWidget::UnivGLWidget(CineSpace *c)
{
    this->setWindowTitle("Vue 3D système solaire");
    cine = c;

    // Génération des éléments du menu contextuel
    setContextMenuPolicy(Qt::ActionsContextMenu);

    //Item pour l'appel de la boite de dialogue
    QAction *confAction = new QAction(tr("&Configuration"), this);
    connect(confAction, SIGNAL(triggered()), this, SLOT(confDialogSlot()));
    addAction(confAction);

    //Item pour le "Help"
    labelHelp = new QLabel(
                tr(
                    "Dans cette vue du système solaire les distances du Soleil aux planètes"
                    "<br/>sont respectées, sauf pour Uranus et Neptune (multipliées par 3/4)."
                    "<br/>Les dimensions des planètes et de la Lune sont multipliées par 500,"
                    "<br/>sauf Uranus et Neptune multipliées par 1000."
                    "<br/>La dimension du Soleil est multipliée par 10."
                    "<h4>Action touches clavier :</h4>"
                    "<ul><li>Flèches haut/bas : Rotation en site.</li>"
                    "<li>Flèches droite/gauche : Rotation en azimut.</li>"
                    "<li>Lettre E/e : Eloignement ou rapprochement.</li>"
                    "<<li>Echappement : Retour point de vue initial.</li>"
                    "</ul>"
                    "<h4>Menu contextuel :</h4>"
                    "<ul><li>Configuration :"
                    "<br/>Affiche une boite de dialogue permettant ?? A faire</li>"
                    "<li>Sauve Image :"
                    "<br/>Permet de sauvegarder la vue de la fenêtre en image jpeg.</li>"
                    "<li>Help :"
                    "<br/>Affiche cette aide.</li>"
                    "</ul>"
                    ));
    labelHelp->setWindowTitle("Aide Vue Système Solaire");

    // Le Slot sauver image
    QAction *sauveImg = new QAction(tr("Sauve Image"), this);
    connect(sauveImg, SIGNAL(triggered()), this, SLOT(grabImageSlot()));
    addAction(sauveImg);

    QAction *actionHelp = new QAction(tr("&Help"), this);
    connect(actionHelp, SIGNAL(triggered()), this, SLOT(afficheHelpSlot()));
    addAction(actionHelp);

    // Format de la fenêtre OpenGL
    setFormat(QGLFormat(QGL::DoubleBuffer | QGL::DepthBuffer | QGL::Rgba));

    // Initialisation paramètre de la scene dans le repère openGL initial
    perspective = 1;            // Flag utilisation perspective
    //    AngleVueVertical = 30. ;    // Champ de vue en degrés
    gainDist = 1.0;             // Pour éloigner ou approcher le point de vue (déplacement camera)
    gainAngle = 1.0;             // Pour éloigner ou approcher le point de vue (zoom)
    sitScene = sitSceneIni = 30.0; // Pour faire bouger la scene en site
    aziScene = aziSceneIni = 0.0; // Pour faire bouger la scene en azimut

    //  On fabrique la boite de dialogue de configuration de la scene à afficher
    //    confEarthDialog = new ConfEarthDialog(this, this);
    terlun = 50.*Lune::dist0/Terre::UA;
}

// Enregistrement d'image-ecran sous format jpg
void UnivGLWidget::grabImageSlot()
{
    QImage image = grabFrameBuffer();
    QString filename = QFileDialog::getSaveFileName
            (this,tr("Sauve Image"),".",tr("Fichier image (*.jpg)"));
    if (!filename.isEmpty()) image.save(filename);
}

// Reception clic contextuel pour l'appel boite de dialigue configuration
void UnivGLWidget::confDialogSlot()
{
    /*
    confEarthDialog->setModal(false);
    confEarthDialog->show();
*/
}

// Dimension par defaut de la fenetre OpenGl
QSize UnivGLWidget::sizeHint() const
{
    return QSize(1024,512);
}

// Si fermeture on fait le ménage
void UnivGLWidget::closeEvent(QCloseEvent *event)
{
    //    if(confEarthDialog) confEarthDialog->close();
    labelHelp->close();
}


// Initialisation OpenGL
void UnivGLWidget::initializeGL()
{
    //  Couleur d'effacement sombre
    glClearColor(0.05, 0.05, 0.05, 1.0);
    glEnable(GL_DEPTH_TEST); // Elimination des faces cachees
    glFrontFace(GL_CCW); // Les faces avant sont celles qui sont parcourues dans le sens direct (par défaut a priori)
    glEnable(GL_CULL_FACE); // Les faces avant ou arrières seront pré-éliminées
    glCullFace (GL_BACK) ; // Ce sont les Faces arrières qui seront pré-éliminées (par défaut a priori)

    // Définition des composantes de la lumiere GL_LIGHT0 utilisée comme lumière solaire
    glLightfv(GL_LIGHT0, GL_AMBIENT, eteinte);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, puissante);
    glLightfv(GL_LIGHT0, GL_SPECULAR, moyenne);
    glEnable(GL_LIGHT0);

    // Amélioration des calculs de l'éclairage
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, tenue); // Lulière ambiante globale atténuée pour faire la nuit coté nuit.
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR); // Pour améliorer le rendu des textures
    glShadeModel(GL_SMOOTH); // Pour un degradé des couleurs (inutile ? car par défaut)

    // Fabrication du Soleil et des planètes texturées dans des DisplayListes
    dl_Soleil = newDlPanete((float) 10.*6.96E8/Terre::UA, 12, ":/leSoleil");
    dl_Lune = newDlPanete((float) 500.*Lune::rayLune/Terre::UA, 24, ":/laLune");
    dl_Planetes[0] = newDlPanete((float) 500.*2.44E6/Terre::UA, 12, ":/Mercure");
    dl_Planetes[1] = newDlPanete((float) 500.*6.05E6/Terre::UA, 12, ":/Venus");
    dl_Planetes[2] = newDlPanete((float) 500.*6.37E6/Terre::UA, 12, ":/laTerre");
    dl_Planetes[3] = newDlPanete((float) 500.*3.39E6/Terre::UA, 12, ":/Mars");
    dl_Planetes[4] = newDlPanete((float) 500.*69.9E6/Terre::UA, 12, ":/Jupiter");
    dl_Planetes[5] = newDlSphere((float) 500.*58.2E6/Terre::UA, 12);
    dl_Planetes[6] = newDlSphere((float) 1000.*25.4E6/Terre::UA, 12);
    dl_Planetes[7] = newDlSphere((float) 1000.*24.6E6/Terre::UA, 12);
    // Fabrication des orbites
    double gain = 1;
    for (int i = 0; i <8; i++)
    {
        if (i >= 6) gain = 0.75;
        dl_Orbs[i] = glNewOrbite(demgaxe[i]*gain, excen[i], inclin[i], gOmega[i], pomega[i], 60);
    }
}

// Si redimmensionnement de la fenêtre, on recalcule les paramètres de projection
void UnivGLWidget::resizeGL(int w, int h)
{
    width1 = w;  height1 = h;
    glViewport(0,0,width1,height1); // On affiche sur toute la fenêtre
    setProjection();
}

/*----------------------------------------------------------------------------------------------
 * Préparation des matrices de projection et scène. On prend en compte les paramètres suivants :
 *   taille de la fenêtre : width1 et height1
 *   modePtVue : le mode de point de vue désiré
 *   AngleVueIni : si grand, effet grand angle, si petit effet zoom.
 *   gainAngle : Coefficient multiplcateur de cet angle
 *   Perspective : booleen précisant si proj avec perspective, ou sans ( orthogonale)
 *   ditSceneIni : distance entre l'observateur et le repère scène
 *   gainDist : coef multiplicateur sur cette distance
 *   aziScene, sitScene : direction du regard de l'observateur
 *--------------------------------------------------------------------------------------*/
void UnivGLWidget::setProjection()
{
    double zProche, zInfini;
    AngleVueIni = 45.;
    distSceneIni = 3.; // Point de vue 3 distance Terre-Soleil
    zProche = 0.0001;
    zInfini = 100000.;

    // Définition de la matrice de projection initale (projetant la vue da
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity(); // modifiée ci-après en fonction de l'angle de vue et du mode perspective ou orthogonal
    if(perspective)
        gluPerspective(gainAngle*AngleVueIni, 1.*width1/height1, zProche, zInfini);
    else
    {
        double Dy = gainDist*distSceneIni*tan(0.5*gainAngle*AngleVueIni*M_PI/180.);
        double Dx = Dy*width1/height1;
        glOrtho(-Dx, Dx, -Dy, Dy, zProche, zInfini);
    }
    // Définition de la matrice transformant le repère observateur (OpenGL) en repère scène
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix(); // Pour eviter de saturer la pile avec le push à la fin, si plusieurs appels consécutifs
    glLoadIdentity(); // modifiée ci-après en fonction de sa situation par rapport au rep. observateur
    glTranslated(0., 0., -gainDist * distSceneIni); // Eloigne la scene de l'oeil (Z est vers l'avant ds Rep openGL)

    // On oriente le repère scène initial en fonction du cas traité
    //--------------------------------------------------------------
    glRotated(-90.+sitScene, 1.0, 0.0, 0.0); // Ramène Z vertical et le penche d'un angle site
    glRotated(-90.-aziScene, 0.0, 0.0, 1.0); // Ramène X vers l'avant et fait tourner la scene en azimut.
    glPushMatrix();
}

void UnivGLWidget::paintGL()
{
    double x,y,z;
    if (height1 == 0) return ;

    // Initialisations
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // On efface ce qu'il y avait avant
    glEnable(GL_LIGHTING);
    glPopMatrix(); // Recupère le repère scène de départ
    glPushMatrix(); // et la sauvegarde pour le prochain appel
    glTranslated(xScene, yScene, zScene); // Pour modif avec touches X, Y, Z et verifier les axes de reference

    glColor3fv(puissante);
    glDisable(GL_COLOR_MATERIAL); // Desactivation du coloriage simple (color tracking)

    // Activation de l'éclairage
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE); // On ne traite que les faces externes
    glColorMaterial(GL_FRONT, GL_DIFFUSE);
    glMaterialfv(GL_FRONT, GL_AMBIENT, eteinte);
    glMaterialfv(GL_FRONT, GL_SPECULAR, moyenne);
    glMateriali(GL_FRONT, GL_SHININESS, 32); // Pour limiter les défauts à la limite eclairage/ombre [0-128] en mode specular
    GLfloat pos_lumiere0[4]; // Pour la direction du Soleil (x, y, z, s avec s=0 pour direction et 1 pour position)

    for (int i=0; i<3; i++) pos_lumiere0[i] = 0.; // Position Soleil au centre
    pos_lumiere0[3] = 1.; // 1. pour position, 0. pour direction
    glCallList(dl_Soleil);
    double gain = 1.;
    for (int i = 0; i < 8; i++)
    {
        if (i >= 6) gain = 0.75;
        glPushMatrix();
        x = Planets::v8x6_OP_e[i][0];
        y = Planets::v8x6_OP_e[i][1];
        z = Planets::v8x6_OP_e[i][2];
        glTranslated(x*gain, y*gain, z*gain);
        if(i == 2)
        {
            double mat44[16], m3[9];
            /* On ajoute la Lune */
            glPushMatrix();
            glTranslated(Lune::v3_uTL_e[0]*terlun, Lune::v3_uTL_e[1]*terlun, Lune::v3_uTL_e[2]*terlun);
            // On oriente la Lune
            tm3m3(m3,Reperes::m3_CE, Lune::m3_CM);
            m3ToGLmat44(mat44, m3); // Convertit m3 en 4x4
            glMultMatrixd(mat44); // On est dans Rm
            glCallList(dl_Lune);
            glPopMatrix();
            // On oriente la Terre
            tm3m3(m3,Reperes::m3_CE,Terre::m3_CG);
            m3ToGLmat44(mat44, m3); // Convertit m3_EG en 4x4
            glMultMatrixd(mat44); // On est dans Rg (repère de greenwich)
        }
        if (i > 4)
        {
            glEnable(GL_COLOR_MATERIAL);
            glColor3dv(colOrb[i]);
        }
        glCallList(dl_Planetes[i]);
        glPopMatrix();
    }
    glDisable(GL_LIGHTING); // couleur indépendante de l'éclairage
    for (int i = 0; i < 8; i++)
    {
        glColor3dv(colOrb[i]);
        glCallList(dl_Orbs[i]);
    }
    return;
}


/*----------------------------------------------------------------------------------*/
/* Callback traitement fleches pour modifier le point de vue                        */
/* ^ - V               : +/- angle Site                                             */
/* > - <               : +/- angle Azimut                                           */
/* Home - End          : fenetre vers la gauche - droite                            */
/* Page-Up - Page-Down : fenetre vers le haut, bas                                  */
/* e - E               : eloigne ou approche l'observateur                          */
/* a - A               : effet de grand angle ou zoom                               */
/* Esc                 : point de vue initial                                       */
/* p, P                : Vue avec ou sans perspective  (choix supprimé)             */
/* x,X,y,Y,z,Z : modifie la position de la scene                                    */
/* 0 ou o ou O : Retour à la configuration initiale                                 */
/*----------------------------------------------------------------------------------*/
void UnivGLWidget::keyPressEvent(QKeyEvent *event)
{
    int key = event->key();
    bool ok = false;
    bool majuscule = (Qt::SHIFT == event->modifiers());
    int maj = majuscule ? 1 : 0;
    int smaj = -1 + 2*maj; /* normal = -1; majuscule = +1 */
    double DeltaA = 0.5, DeltaL = distSceneIni/100.;

    switch (key)
    {
    case Qt::Key_Up :
        sitScene += DeltaA ;
        if (sitScene > 180.) sitScene -= 360. ;
        ok = true;
        break ;
    case Qt::Key_Down :
        sitScene -= DeltaA ;
        if (sitScene < -180.) sitScene += 360. ;
        ok = true;
        break ;
    case Qt::Key_Left :
        aziScene -= DeltaA ;
        if (aziScene < -180.) aziScene += 360. ;
        ok = true;
        break ;
    case Qt::Key_Right :
        aziScene += DeltaA ;
        if (aziScene > 180.) aziScene -= 360. ;
        ok = true;
        break ;
    case Qt::Key_Escape : /* Point de vue initial */
        aziScene = aziSceneIni;
        sitScene = sitSceneIni;
        gainDist = 1.;
        gainAngle = 1.;
        xScene = yScene = zScene = 0.;
        ok = true;
        break ;
    case Qt::Key_E : /* gain sur distScene */
        if (majuscule) gainDist *= 0.975;
        else gainDist /= 0.975;
        ok = true;
        break ;
    case Qt::Key_A : /* gain sur angle de vue */
        if (majuscule) gainAngle *= 0.975;
        else gainAngle /= 0.975;
        ok = true;
        break ;
    case Qt::Key_X : /* +/- X */
        xScene -= DeltaL*smaj ;
        ok = true;
        break ;
    case Qt::Key_Y : /* +/- Y */
        yScene -= DeltaL*smaj ;
        ok = true;
        break ;
    case Qt::Key_Z : /* +/- Z */
        zScene -= DeltaL*smaj ;
        ok = true;
        break ;

    case Qt::Key_P : /* Perspective */
        perspective = !maj ;
        ok = true;
        break;

    }
    if (ok)
    {
        setProjection();
        updateGL();
    }
    else QGLWidget::keyPressEvent(event);
}
